import json
from typing import Any, Optional, Union

import humps
# from composio.constants import DEFAULT_ENTITY_ID
from pydantic import BaseModel

from mirix.constants import COMPOSIO_ENTITY_ENV_VAR_KEY, DEFAULT_MESSAGE_TOOL, DEFAULT_MESSAGE_TOOL_KWARG
from mirix.schemas.enums import MessageRole
from mirix.schemas.mirix_message import AssistantMessage, ReasoningMessage, ToolCallMessage
from mirix.schemas.mirix_response import MirixResponse
from mirix.schemas.message import MessageCreate

def generate_langchain_tool_wrapper(
    tool: "LangChainBaseTool", additional_imports_module_attr_map: dict[str, str] = None
) -> tuple[str, str]:
    tool_name = tool.__class__.__name__
    import_statement = f"from langchain_community.tools import {tool_name}"
    extra_module_imports = generate_import_code(additional_imports_module_attr_map)

    # Safety check that user has passed in all required imports:
    assert_all_classes_are_imported(tool, additional_imports_module_attr_map)

    tool_instantiation = f"tool = {generate_imported_tool_instantiation_call_str(tool)}"
    run_call = f"return tool._run(**kwargs)"
    func_name = humps.decamelize(tool_name)

    # Combine all parts into the wrapper function
    wrapper_function_str = f"""
def {func_name}(**kwargs):
    import importlib
    {import_statement}
    {extra_module_imports}
    {tool_instantiation}
    {run_call}
"""

    # Compile safety check
    assert_code_gen_compilable(wrapper_function_str)

    return func_name, wrapper_function_str


def assert_code_gen_compilable(code_str):
    try:
        compile(code_str, "<string>", "exec")
    except SyntaxError as e:
        print(f"Syntax error in code: {e}")


def assert_all_classes_are_imported(tool: Union["LangChainBaseTool"], additional_imports_module_attr_map: dict[str, str]) -> None:
    # Safety check that user has passed in all required imports:
    tool_name = tool.__class__.__name__
    current_class_imports = {tool_name}
    if additional_imports_module_attr_map:
        current_class_imports.update(set(additional_imports_module_attr_map.values()))
    required_class_imports = set(find_required_class_names_for_import(tool))

    if not current_class_imports.issuperset(required_class_imports):
        err_msg = f"[ERROR] You are missing module_attr pairs in `additional_imports_module_attr_map`. Currently, you have imports for {current_class_imports}, but the required classes for import are {required_class_imports}"
        print(err_msg)
        raise RuntimeError(err_msg)


def find_required_class_names_for_import(obj: Union["LangChainBaseTool", BaseModel]) -> list[str]:
    """
    Finds all the class names for required imports when instantiating the `obj`.
    NOTE: This does not return the full import path, only the class name.

    We accomplish this by running BFS and deep searching all the BaseModel objects in the obj parameters.
    """
    class_names = {obj.__class__.__name__}
    queue = [obj]

    while queue:
        # Get the current object we are inspecting
        curr_obj = queue.pop()

        # Collect all possible candidates for BaseModel objects
        candidates = []
        if is_base_model(curr_obj):
            # If it is a base model, we get all the values of the object parameters
            # i.e., if obj('b' = <class A>), we would want to inspect <class A>
            fields = dict(curr_obj)
            # Generate code for each field, skipping empty or None values
            candidates = list(fields.values())
        elif isinstance(curr_obj, dict):
            # If it is a dictionary, we get all the values
            # i.e., if obj = {'a': 3, 'b': <class A>}, we would want to inspect <class A>
            candidates = list(curr_obj.values())
        elif isinstance(curr_obj, list):
            # If it is a list, we inspect all the items in the list
            # i.e., if obj = ['a', 3, None, <class A>], we would want to inspect <class A>
            candidates = curr_obj

        # Filter out all candidates that are not BaseModels
        # In the list example above, ['a', 3, None, <class A>], we want to filter out 'a', 3, and None
        candidates = filter(lambda x: is_base_model(x), candidates)

        # Classic BFS here
        for c in candidates:
            c_name = c.__class__.__name__
            if c_name not in class_names:
                class_names.add(c_name)
                queue.append(c)

    return list(class_names)


def generate_imported_tool_instantiation_call_str(obj: Any) -> Optional[str]:
    if isinstance(obj, (int, float, str, bool, type(None))):
        # This is the base case
        # If it is a basic Python type, we trivially return the string version of that value
        # Handle basic types
        return repr(obj)
    elif is_base_model(obj):
        # Otherwise, if it is a BaseModel
        # We want to pull out all the parameters, and reformat them into strings
        # e.g. {arg}={value}
        # The reason why this is recursive, is because the value can be another BaseModel that we need to stringify
        model_name = obj.__class__.__name__
        fields = obj.dict()
        # Generate code for each field, skipping empty or None values
        field_assignments = []
        for arg, value in fields.items():
            python_string = generate_imported_tool_instantiation_call_str(value)
            if python_string:
                field_assignments.append(f"{arg}={python_string}")

        assignments = ", ".join(field_assignments)
        return f"{model_name}({assignments})"
    elif isinstance(obj, dict):
        # Inspect each of the items in the dict and stringify them
        # This is important because the dictionary may contain other BaseModels
        dict_items = []
        for k, v in obj.items():
            python_string = generate_imported_tool_instantiation_call_str(v)
            if python_string:
                dict_items.append(f"{repr(k)}: {python_string}")

        joined_items = ", ".join(dict_items)
        return f"{{{joined_items}}}"
    elif isinstance(obj, list):
        # Inspect each of the items in the list and stringify them
        # This is important because the list may contain other BaseModels
        list_items = [generate_imported_tool_instantiation_call_str(v) for v in obj]
        filtered_list_items = list(filter(None, list_items))
        list_items = ", ".join(filtered_list_items)
        return f"[{list_items}]"
    else:
        # Otherwise, if it is none of the above, that usually means it is a custom Python class that is NOT a BaseModel
        # Thus, we cannot get enough information about it to stringify it
        # This may cause issues, but we are making the assumption that any of these custom Python types are handled correctly by the parent library, such as LangChain
        # An example would be that WikipediaAPIWrapper has an argument that is a wikipedia (pip install wikipedia) object
        # We cannot stringify this easily, but WikipediaAPIWrapper handles the setting of this parameter internally
        # This assumption seems fair to me, since usually they are external imports, and LangChain should be bundling those as module-level imports within the tool
        # We throw a warning here anyway and provide the class name
        print(
            f"[WARNING] Skipping parsing unknown class {obj.__class__.__name__} (does not inherit from the Pydantic BaseModel and is not a basic Python type)"
        )
        if obj.__class__.__name__ == "function":
            import inspect

            print(inspect.getsource(obj))

        return None


def is_base_model(obj: Any):
    from langchain_core.pydantic_v1 import BaseModel as LangChainBaseModel

    return isinstance(obj, BaseModel) or isinstance(obj, LangChainBaseModel)


def generate_import_code(module_attr_map: Optional[dict]):
    if not module_attr_map:
        return ""

    code_lines = []
    for module, attr in module_attr_map.items():
        module_name = module.split(".")[-1]
        code_lines.append(f"# Load the module\n    {module_name} = importlib.import_module('{module}')")
        code_lines.append(f"    # Access the {attr} from the module")
        code_lines.append(f"    {attr} = getattr({module_name}, '{attr}')")
    return "\n".join(code_lines)


def parse_mirix_response_for_assistant_message(
    mirix_response: MirixResponse,
    assistant_message_tool_name: str = DEFAULT_MESSAGE_TOOL,
    assistant_message_tool_kwarg: str = DEFAULT_MESSAGE_TOOL_KWARG,
) -> Optional[str]:
    reasoning_message = ""
    for m in mirix_response.messages:
        if isinstance(m, AssistantMessage):
            return m.assistant_message
        elif isinstance(m, ToolCallMessage) and m.tool_call.name == assistant_message_tool_name:
            try:
                return json.loads(m.tool_call.arguments)[assistant_message_tool_kwarg]
            except Exception:  # TODO: Make this more specific
                continue
        elif isinstance(m, ReasoningMessage):
            # This is not ideal, but we would like to return something rather than nothing
            reasoning_message += f"{m.reasoning}\n"

    return None


import asyncio
from random import uniform
from typing import Optional


async def async_send_message_with_retries(
    server,
    sender_agent: "Agent",
    target_agent_id: str,
    message_text: str,
    max_retries: int,
    timeout: int,
    logging_prefix: Optional[str] = None,
) -> str:
    """
    Shared helper coroutine to send a message to an agent with retries and a timeout.

    Args:
        server: The Mirix server instance (from get_mirix_server()).
        sender_agent (Agent): The agent initiating the send action.
        target_agent_id (str): The ID of the agent to send the message to.
        message_text (str): The text to send as the user message.
        max_retries (int): Maximum number of retries for the request.
        timeout (int): Maximum time to wait for a response (in seconds).
        logging_prefix (str): A prefix to append to logging
    Returns:
        str: The response or an error message.
    """
    logging_prefix = logging_prefix or "[async_send_message_with_retries]"
    for attempt in range(1, max_retries + 1):
        try:
            messages = [MessageCreate(role=MessageRole.user, text=message_text, name=sender_agent.agent_state.name)]
            # Wrap in a timeout
            response = await asyncio.wait_for(
                server.send_message_to_agent(
                    agent_id=target_agent_id,
                    actor=sender_agent.user,
                    messages=messages,
                    stream_steps=False,
                    stream_tokens=False,
                    use_assistant_message=True,
                    assistant_message_tool_name=DEFAULT_MESSAGE_TOOL,
                    assistant_message_tool_kwarg=DEFAULT_MESSAGE_TOOL_KWARG,
                ),
                timeout=timeout,
            )

            # Extract assistant message
            assistant_message = parse_mirix_response_for_assistant_message(
                response,
                assistant_message_tool_name=DEFAULT_MESSAGE_TOOL,
                assistant_message_tool_kwarg=DEFAULT_MESSAGE_TOOL_KWARG,
            )
            if assistant_message:
                msg = f"Agent {target_agent_id} said '{assistant_message}'"
                sender_agent.logger.info(f"{logging_prefix} - {msg}")
                return msg
            else:
                msg = f"(No response from agent {target_agent_id})"
                sender_agent.logger.info(f"{logging_prefix} - {msg}")
                return msg
        except asyncio.TimeoutError:
            error_msg = f"(Timeout on attempt {attempt}/{max_retries} for agent {target_agent_id})"
            sender_agent.logger.warning(f"{logging_prefix} - {error_msg}")
        except Exception as e:
            error_msg = f"(Error on attempt {attempt}/{max_retries} for agent {target_agent_id}: {e})"
            sender_agent.logger.warning(f"{logging_prefix} - {error_msg}")

        # Exponential backoff before retrying
        if attempt < max_retries:
            backoff = uniform(0.5, 2) * (2**attempt)
            sender_agent.logger.warning(f"{logging_prefix} - Retrying the agent to agent send_message...sleeping for {backoff}")
            await asyncio.sleep(backoff)
        else:
            sender_agent.logger.error(f"{logging_prefix} - Fatal error during agent to agent send_message: {error_msg}")
            return error_msg



from mirix import system
from mirix.schemas.enums import MessageRole
from mirix.schemas.mirix_message_content import TextContent, ImageContent, FileContent, CloudFileContent
from mirix.schemas.message import Message, MessageCreate


def extract_and_wrap_message_content(
                                     role: MessageRole,
                                     content: str | TextContent | ImageContent | FileContent | CloudFileContent,
                                     wrap_user_message: bool = False,
                                     wrap_system_message: bool = True,):
    
    # Extract message content
    if isinstance(content, str):
        message_content = content
    elif content and isinstance(content, TextContent):
        message_content = content.text
    elif content and isinstance(content, ImageContent):
        # Skip wrapping if the content is an image
        return content
    elif content and isinstance(content, FileContent):
        # Skip wrapping if the content is a file
        return content
    elif content and isinstance(content, CloudFileContent):
        return content
    else:
        raise ValueError("Message content is empty or invalid")

    # Apply wrapping if needed
    if role == MessageRole.user and wrap_user_message:
        message_content = system.package_user_message(user_message=message_content)
    elif role == MessageRole.system and wrap_system_message:
        message_content = system.package_system_message(system_message=message_content)
    elif role not in {MessageRole.user, MessageRole.system}:
        raise ValueError(f"Invalid message role: {role}")
    return TextContent(text=message_content)

def prepare_input_message_create(
    message: MessageCreate,
    agent_id: str,
    **kwargs
) -> Message:
    """Converts a MessageCreate object into a Message object, applying wrapping if needed."""
    # TODO: This seems like extra boilerplate with little benefit
    assert isinstance(message, MessageCreate)

    if isinstance(message.content, list):
        content = [extract_and_wrap_message_content(role=message.role, content=c, **kwargs) for c in message.content]
    else:
        content = [extract_and_wrap_message_content(role=message.role, content=message.content, **kwargs)]

    return Message(
        agent_id=agent_id,
        role=message.role,
        content=content,
        name=message.name,
        model=None,  # assigned later?
        tool_calls=None,  # irrelevant
        tool_call_id=None,
        otid=message.otid,
        sender_id=message.sender_id,
        group_id=message.group_id,
    )



from datetime import datetime
from typing import Dict, List, Optional, Any

from pydantic import Field, field_validator
from mirix.constants import MAX_EMBEDDING_DIM

from mirix.schemas.mirix_base import MirixBase
from mirix.schemas.embedding_config import EmbeddingConfig
from mirix.utils import get_utc_time

class ResourceMemoryItemBase(MirixBase):
    """
    Base schema for resource memory items - storing docs, user files, references, etc.
    """
    __id_prefix__ = "res_item"
    title: str = Field(..., description="Short name/title of the resource")
    summary: str = Field(..., description="Short description or summary of the resource")
    resource_type: str = Field(..., description="File type or format (e.g. 'doc', 'markdown', 'pdf_text')")
    content: str = Field(..., description="Full or partial text content of the resource")
    tree_path: List[str] = Field(..., description="Hierarchical categorization path as an array of strings (e.g., ['documents', 'work', 'projects'])")

class ResourceMemoryItem(ResourceMemoryItemBase):
    """
    Full schema for resource memory items with DB fields.
    """
    id: Optional[str] = Field(None, description="Unique identifier for the resource memory item")
    user_id: str = Field(..., description="The id of the user who generated the resource")
    created_at: datetime = Field(default_factory=get_utc_time, description="Creation timestamp")
    updated_at: Optional[datetime] = Field(None, description="Last update timestamp")
    last_modify: Dict[str, Any] = Field(
        default_factory=lambda: {"timestamp": get_utc_time().isoformat(), "operation": "created"},
        description="Last modification info including timestamp and operation type"
    )
    organization_id: str = Field(..., description="The unique identifier of the organization")
    summary_embedding: Optional[List[float]] = Field(None, description="The embedding of the summary")
    embedding_config: Optional[EmbeddingConfig] = Field(None, description="The embedding configuration used by the event")
    metadata_: Dict[str, Any] = Field(default_factory=dict, description="Arbitrary additional metadata (tags, creation date, etc.)")

    @field_validator("summary_embedding")
    @classmethod
    def pad_embeddings(cls, embedding: List[float]) -> List[float]:
        """Pad embeddings to `MAX_EMBEDDING_SIZE`. This is necessary to ensure all stored embeddings are the same size."""
        import numpy as np

        if embedding and len(embedding) != MAX_EMBEDDING_DIM:
            np_embedding = np.array(embedding)
            padded_embedding = np.pad(np_embedding, (0, MAX_EMBEDDING_DIM - np_embedding.shape[0]), mode="constant")
            return padded_embedding.tolist()
        return embedding

class ResourceMemoryItemUpdate(MirixBase):
    """Schema for updating an existing resource memory item."""
    id: str = Field(..., description="Unique ID for this resource memory entry")
    title: Optional[str] = Field(None, description="Short name/title of the resource")
    summary: Optional[str] = Field(None, description="Short description or summary of the resource")
    resource_type: Optional[str] = Field(None, description="File type/format (e.g. 'doc', 'markdown')")
    content: Optional[str] = Field(None, description="Full or partial text content")
    tree_path: Optional[List[str]] = Field(
        None, 
        description="Hierarchical categorization path as an array of strings (e.g., ['documents', 'work', 'projects'])"
    )
    organization_id: Optional[str] = Field(None, description="The organization ID")
    updated_at: datetime = Field(default_factory=get_utc_time, description="Update timestamp")
    last_modify: Optional[Dict[str, Any]] = Field(
        None,
        description="Last modification info including timestamp and operation type"
    )
    summary_embedding: Optional[List[float]] = Field(None, description="The embedding of the summary")
    embedding_config: Optional[EmbeddingConfig] = Field(None, description="The embedding configuration used by the event")
    metadata_: Optional[Dict[str, Any]] = Field(None, description="Arbitrary additional metadata")

class ResourceMemoryItemResponse(ResourceMemoryItem):
    """Response schema for resource memory item with additional fields if needed."""
    pass



from typing import List, Optional

from mirix.llm_api.llm_client_base import LLMClientBase
from mirix.schemas.llm_config import LLMConfig
from mirix.schemas.message import Message
from mirix.schemas.openai.chat_completion_response import ChatCompletionResponse


class CohereClient(LLMClientBase):
    """Cohere client - currently a stub that falls back to original implementation"""

    def build_request_data(
        self,
        messages: List[Message],
        llm_config: LLMConfig,
        tools: Optional[List[dict]] = None,
        force_tool_call: Optional[str] = None,
    ) -> dict:
        # TODO: Implement cohere-specific request building
        raise NotImplementedError("CohereClient not yet implemented - use fallback")

    def request(self, request_data: dict) -> dict:
        # TODO: Implement cohere-specific request
        raise NotImplementedError("CohereClient not yet implemented - use fallback")

    def convert_response_to_chat_completion(
        self,
        response_data: dict,
        input_messages: List[Message],
    ) -> ChatCompletionResponse:
        # TODO: Implement cohere-specific response conversion
        raise NotImplementedError("CohereClient not yet implemented - use fallback")

    def handle_llm_error(self, e: Exception) -> Exception:
        # TODO: Implement cohere-specific error handling
        return super().handle_llm_error(e) 


from datetime import datetime
from typing import Dict, List, Optional, Any

from pydantic import Field, field_validator
from mirix.constants import MAX_EMBEDDING_DIM

from mirix.schemas.mirix_base import MirixBase
from mirix.schemas.embedding_config import EmbeddingConfig
from mirix.utils import get_utc_time


class EpisodicEventBase(MirixBase):
    """
    Base schema for episodic memory events containing common fields.
    """

    __id_prefix__ = "ep_mem"
    event_type: str = Field(..., description="Type/category of the episodic event (e.g., user_message, inference, system_notification)")
    summary: str = Field(..., description="Short textual summary of the event")
    details: str = Field(..., description="Detailed description or text for the event")
    actor: str = Field(..., description="The actor who generated the event (user or assistant)")
    tree_path: List[str] = Field(..., description="Hierarchical categorization path as an array of strings (e.g., ['work', 'projects', 'ai-research'])")

class EpisodicEventForLLM(EpisodicEventBase):
    """
    Schema for creating a new episodic memory record.
    """
    # TODO: make `occurred_at` optional
    occurred_at: str = Field(..., description="When the event happened (it should be mentioned in the user's response and it should be in the format of 'YYYY-MM-DD HH:MM:SS')")

class EpisodicEvent(EpisodicEventBase):
    """
    Representation of a single episodic memory event in the system.

    Additional Parameters:
        id (str): Unique identifier for this memory item
        occurred_at (datetime): When the event occurred or was recorded
        created_at (datetime): When the memory record was created in the system
        updated_at (Optional[datetime]): Last update timestamp
    """
    id: Optional[str] = Field(None, description="Unique identifier for the episodic event")

    user_id: str = Field(..., description="The id of the user who generated the episodic event")

    occurred_at: datetime = Field(
        default_factory=get_utc_time,
        description="When the event actually happened (recorded or user-labeled)."
    )
    created_at: datetime = Field(
        default_factory=get_utc_time,
        description="Timestamp when this memory record was created"
    )
    updated_at: Optional[datetime] = Field(
        None,
        description="When this memory record was last updated"
    )
    last_modify: Dict[str, Any] = Field(
        default_factory=lambda: {"timestamp": get_utc_time().isoformat(), "operation": "created"},
        description="Last modification info including timestamp and operation type"
    )
    metadata_: Dict[str, Any] = Field(default_factory=dict, description="Additional structured metadata for the event")
    organization_id: str = Field(..., description="Unique identifier of the organization")
    details_embedding: Optional[List[float]] = Field(None, description="The embedding of the event")
    summary_embedding: Optional[List[float]] = Field(None, description="The embedding of the summary")
    embedding_config: Optional[EmbeddingConfig] = Field(None, description="The embedding configuration used by the event")

    # need to validate both details_embedding and summary_embedding to ensure they are the same size
    @field_validator("details_embedding", "summary_embedding")
    @classmethod
    def pad_embeddings(cls, embedding: List[float]) -> List[float]:
        """Pad embeddings to `MAX_EMBEDDING_SIZE`. This is necessary to ensure all stored embeddings are the same size."""
        import numpy as np

        if embedding and len(embedding) != MAX_EMBEDDING_DIM:
            np_embedding = np.array(embedding)
            padded_embedding = np.pad(np_embedding, (0, MAX_EMBEDDING_DIM - np_embedding.shape[0]), mode="constant")
            return padded_embedding.tolist()
        return embedding

class EpisodicEventUpdate(MirixBase):
    """
    Schema for updating an existing episodic memory record.

    All fields (except id) are optional so that only provided fields are updated.
    """
    id: str = Field(..., description="Unique ID for this episodic memory record")
    event_type: Optional[str] = Field(None, description="Type/category of the event")
    summary: Optional[str] = Field(None, description="Short textual summary of the event")
    details: Optional[str] = Field(None, description="Detailed text describing the event")
    tree_path: Optional[List[str]] = Field(
        None, 
        description="Hierarchical categorization path as an array of strings (e.g., ['work', 'projects', 'ai-research'])"
    )
    metadata_: Optional[Dict[str, Any]] = Field(None, description="Any additional metadata")
    organization_id: Optional[str] = Field(None, description="Unique identifier of the organization")
    occurred_at: Optional[datetime] = Field(None, description="If the event's time is updated")
    updated_at: datetime = Field(
        default_factory=get_utc_time,
        description="Timestamp when this memory record was last updated"
    )
    last_modify: Optional[Dict[str, Any]] = Field(
        None,
        description="Last modification info including timestamp and operation type"
    )
    summary_embedding: Optional[List[float]] = Field(None, description="The embedding of the summary")
    details_embedding: Optional[List[float]] = Field(None, description="The embedding of the event")
    embedding_config: Optional[EmbeddingConfig] = Field(None, description="The embedding configuration used by the event")




import os
from mirix import Mirix
from dotenv import load_dotenv

load_dotenv()

API_KEY = os.getenv("AZURE_OPENAI_API_KEY")

assistant = Mirix(
  api_key=API_KEY,
  model_provider="azure_opena",
  config_path=".local/mirix_jplml_azure.yaml",
)

out = assistant.add("Remember me this week's schedule：Monday–Friday (Workdays): 08:30–09:00: Arrive at work/home office, check emails, review calendar, plan tasks for the day. 09:00–10:30: Coding session (feature development, bug fixes, code reviews). 10:30–10:45: Short break (coffee, walk, chat with colleagues). 10:45–12:00: Continue coding, attend stand-up meeting, sync with team. 12:00–13:00: Lunch break. 13:00–15:00: Coding session (implementing new features, writing tests). 15:00–15:15: Break (stretch, snack). 15:15–17:00: Pair programming, debugging, or technical discussion. 17:00–17:30: Wrap up, update tickets, commit/push code, plan for tomorrow. Saturday (Weekend): 09:00–10:00: Leisure (reading, breakfast, light exercise). 10:00–12:00: Personal coding project, open-source contribution, or learning new tech. 12:00–13:00: Lunch. 13:00–15:00: Continue personal project or tech learning. 15:00–18:00: Social activities, hobbies, or relaxation. Sunday (Weekend): 09:00–11:00: Leisure (family time, outdoor activity). 11:00–13:00: Review personal goals, plan for the week, light coding or reading. 13:00–14:00: Lunch. 14:00–18:00: Rest, hobbies, prepare for the upcoming workweek.")

response = assistant.chat("What's my schedule for this week?")

print(response)

# assistant.clear_conversation_history()



from datetime import datetime
from typing import Dict, List, Optional, Any

from pydantic import Field, field_validator
from mirix.constants import MAX_EMBEDDING_DIM

from mirix.schemas.mirix_base import MirixBase
from mirix.utils import get_utc_time
from mirix.schemas.embedding_config import EmbeddingConfig

class SemanticMemoryItemBase(MirixBase):
    """
    Base schema for storing semantic memory items (e.g., general knowledge, concepts, facts).
    """
    __id_prefix__ = "sem_item"
    name: str = Field(..., description="The name or main concept/object for the knowledge entry")
    summary: str = Field(..., description="A concise explanation or summary of the concept")
    details: str = Field(..., description="Detailed explanation or additional context for the concept")
    source: str = Field(..., description="Reference or origin of this information (e.g., book, article, movie)")
    tree_path: List[str] = Field(..., description="Hierarchical categorization path as an array of strings (e.g., ['favorites', 'pets', 'dog'])")

class SemanticMemoryItem(SemanticMemoryItemBase):
    """
    Full semantic memory item schema, including database-related fields.
    """
    id: Optional[str] = Field(None, description="Unique identifier for the semantic memory item")
    user_id: str = Field(..., description="The id of the user who generated the semantic memory")
    created_at: datetime = Field(default_factory=get_utc_time, description="Creation timestamp")
    updated_at: Optional[datetime] = Field(None, description="Last update timestamp")
    last_modify: Dict[str, Any] = Field(
        default_factory=lambda: {"timestamp": get_utc_time().isoformat(), "operation": "created"},
        description="Last modification info including timestamp and operation type"
    )
    metadata_: Dict[str, Any] = Field(default_factory=dict, description="Additional arbitrary metadata as a JSON object")
    organization_id: str = Field(..., description="The unique identifier of the organization")
    details_embedding: Optional[List[float]] = Field(None, description="The embedding of the details")
    name_embedding: Optional[List[float]] = Field(None, description="The embedding of the name")
    summary_embedding: Optional[List[float]] = Field(None, description="The embedding of the summary")
    embedding_config: Optional[EmbeddingConfig] = Field(None, description="The embedding configuration used by the event")

    # need to validate both details_embedding and summary_embedding to ensure they are the same size
    @field_validator("details_embedding", "summary_embedding", "name_embedding")
    @classmethod
    def pad_embeddings(cls, embedding: List[float]) -> List[float]:
        """Pad embeddings to `MAX_EMBEDDING_SIZE`. This is necessary to ensure all stored embeddings are the same size."""
        import numpy as np

        if embedding and len(embedding) != MAX_EMBEDDING_DIM:
            np_embedding = np.array(embedding)
            padded_embedding = np.pad(np_embedding, (0, MAX_EMBEDDING_DIM - np_embedding.shape[0]), mode="constant")
            return padded_embedding.tolist()
        return embedding

class SemanticMemoryItemUpdate(MirixBase):
    """
    Schema for updating an existing semantic memory item.
    """
    id: str = Field(..., description="Unique ID for this semantic memory entry")
    name: Optional[str] = Field(None, description="The name or main concept for the knowledge entry")
    summary: Optional[str] = Field(None, description="A concise explanation or summary of the concept")
    details: Optional[str] = Field(None, description="Detailed explanation or additional context for the concept")
    source: Optional[str] = Field(None, description="Reference or origin of this information (e.g., book, article, movie)")
    actor: Optional[str] = Field(None, description="The actor who generated the semantic memory (user or assistant)")
    tree_path: Optional[List[str]] = Field(
        None, 
        description="Hierarchical categorization path as an array of strings (e.g., ['favorites', 'pets', 'dog'])"
    )
    metadata_: Optional[Dict[str, Any]] = Field(None, description="Additional arbitrary metadata as a JSON object")
    organization_id: Optional[str] = Field(None, description="The organization ID")
    updated_at: datetime = Field(default_factory=get_utc_time, description="Update timestamp")
    last_modify: Optional[Dict[str, Any]] = Field(
        None,
        description="Last modification info including timestamp and operation type"
    )
    details_embedding: Optional[List[float]] = Field(None, description="The embedding of the details")
    name_embedding: Optional[List[float]] = Field(None, description="The embedding of the name")
    summary_embedding: Optional[List[float]] = Field(None, description="The embedding of the summary")
    embedding_config: Optional[EmbeddingConfig] = Field(None, description="The embedding configuration used by the event")


class SemanticMemoryItemResponse(SemanticMemoryItem):
    """
    Response schema for semantic memory item.
    """
    pass



#!/usr/bin/env python3

import os
from setuptools import setup, find_packages

# Read the contents of README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Read requirements from requirements.txt
def parse_requirements(filename):
    """Load requirements from a pip requirements file."""
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

# Filter requirements - exclude some development/system specific packages
def filter_requirements(requirements):
    """Filter out packages that shouldn't be in install_requires."""
    exclude_patterns = [
        'ffmpeg',  # System dependency, not a pip package
    ]
    
    filtered = []
    for req in requirements:
        # Skip empty lines and comments
        if not req or req.startswith('#'):
            continue
        
        # Skip packages that match exclude patterns
        if any(pattern in req.lower() for pattern in exclude_patterns):
            continue
            
        filtered.append(req)
    
    return filtered

# Get version from mirix/__init__.py
def get_version():
    import re
    version_file = os.path.join(this_directory, 'mirix', '__init__.py')
    with open(version_file, 'r', encoding='utf-8') as f:
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M)
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")

requirements = parse_requirements('requirements.txt')
install_requires = filter_requirements(requirements)

setup(
    name='mirix',
    version=get_version(),
    author='Mirix AI',
    author_email='yuwang@mirix.io',
    description='Multi-Agent Personal Assistant with an Advanced Memory System',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Mirix-AI/MIRIX',
    project_urls={
        'Documentation': 'https://docs.mirix.io',
        'Website': 'https://mirix.io',
        'Source Code': 'https://github.com/Mirix-AI/MIRIX',
        'Bug Reports': 'https://github.com/Mirix-AI/MIRIX/issues',
    },
    packages=find_packages(exclude=['tests*', 'scripts*', 'frontend*', 'public_evaluations*']),
    include_package_data=True,
    package_data={
        'mirix': [
            '*.yaml',
            '*.yml', 
            '*.txt',
            'configs/*.yaml',
            'configs/*.yml',
            'prompts/**/*.txt',
            'prompts/**/*.yaml',
            'prompts/**/*.yml',
        ],
    },
    install_requires=install_requires,
    extras_require={
        'dev': [
            'pytest>=6.0.0',
            'pytest-asyncio',
            'black',
            'isort',
            'flake8',
        ],
        'voice': [
            'SpeechRecognition',
            'pydub',
        ],
        'full': [
            'SpeechRecognition',
            'pydub',
        ]
    },
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Communications :: Chat',
    ],
    keywords='ai, memory, agent, llm, assistant, chatbot, multimodal',
    entry_points={
        'console_scripts': [
            'mirix=mirix.__main__:main',
        ],
    },
    license='Apache License 2.0',
    zip_safe=False,
)



"""
Mirix SDK - Simple Python interface for memory-enhanced AI agents
"""

import os
import logging
from typing import Optional, Dict, Any, List, Union
from pathlib import Path

from mirix.agent import AgentWrapper

logger = logging.getLogger(__name__)


class Mirix:
    """
    Simple SDK interface for Mirix memory agent.
    
    Example:
        from mirix import Mirix
        
        memory_agent = Mirix(api_key="your-api-key")
        memory_agent.add("The moon now has a president")
        response = memory_agent.chat("Does moon have a president now?")
    """
    
    def __init__(
        self,
        api_key: str,
        model_provider: str = "google_ai",
        model: Optional[str] = None,
        config_path: Optional[str] = None,
        load_from: Optional[str] = None,
        **kwargs
    ):
        """
        Initialize Mirix memory agent.
        
        Args:
            api_key: API key for LLM provider (required)
            model_provider: LLM provider name (default: "google_ai")
            model: Model to use (optional). If None, uses model from config file.
            config_path: Path to custom config file (optional)
            load_from: Path to backup directory to restore from (optional)
        """
        if not api_key:
            raise ValueError("api_key is required to initialize Mirix")
        
        # Set API key environment variable based on provider
        if model_provider.lower() in ["google", "google_ai", "gemini"]:
            os.environ["GEMINI_API_KEY"] = api_key
        elif model_provider.lower() in ["openai", "gpt"]:
            os.environ["OPENAI_API_KEY"] = api_key
        elif model_provider.lower() in ["anthropic", "claude"]:
            os.environ["ANTHROPIC_API_KEY"] = api_key
        else:
            # For custom providers, use the provider name as prefix
            os.environ[f"{model_provider.upper()}_API_KEY"] = api_key
        
        # Force reload of model_settings to pick up new environment variables
        self._reload_model_settings()
        
        # Track if config_path was originally provided
        config_path_provided = config_path is not None
        
        # Use default config if not specified
        if not config_path:
            # Try to find config file in order of preference
            import sys
            
            if getattr(sys, 'frozen', False):
                # Running in PyInstaller bundle
                bundle_dir = Path(sys._MEIPASS)
                config_path = bundle_dir / "mirix" / "configs" / "mirix.yaml"
                
                if not config_path.exists():
                    raise FileNotFoundError(
                        f"Could not find mirix.yaml config file in PyInstaller bundle at:\n"
                        f"  - {config_path}\n"
                        f"Please ensure config file is properly bundled."
                    )
            else:
                # Running in development - use existing logic
                package_dir = Path(__file__).parent
                
                # 1. Look in package configs directory (for installed package)
                config_path = package_dir / "configs" / "mirix.yaml"
                
                if not config_path.exists():
                    # 2. Look in parent configs directory (for development)
                    config_path = package_dir.parent / "configs" / "mirix.yaml"
                    
                    if not config_path.exists():
                        # 3. Look in current working directory
                        config_path = Path("./mirix/configs/mirix.yaml")
                        
                        if not config_path.exists():
                            raise FileNotFoundError(
                                f"Could not find mirix.yaml config file. Searched in:\n"
                                f"  - {package_dir / 'configs' / 'mirix.yaml'}\n"
                                f"  - {package_dir.parent / 'configs' / 'mirix.yaml'}\n"
                                f"  - {Path('./mirix/configs/mirix.yaml').absolute()}\n"
                                f"Please provide config_path parameter or ensure config file exists."
                            )
        
        # Initialize the underlying agent (with optional backup restore)
        self._agent = AgentWrapper(str(config_path), load_from=load_from)
        
        # Handle model configuration based on parameters:
        # Case 1: model given, config_path None -> load default config then set provided model
        # Case 2: model None, config_path given -> load from config_path and use model from config
        # Case 3: model None, config_path None -> load default config and use default model
        if model is not None:
            # Model explicitly provided - override the config file's model
            self._agent.set_model(model)
            self._agent.set_memory_model(model)
        elif not config_path_provided:
            # No model or config provided - use default model
            default_model = "gemini-2.0-flash"
            self._agent.set_model(default_model)
            self._agent.set_memory_model(default_model)
        # If model is None and config_path was provided, use the model specified in the config file (no override needed)
    
    def add(self, content: str, **kwargs) -> Dict[str, Any]:
        """
        Add information to memory.
        
        Args:
            content: Information to memorize
            **kwargs: Additional options (images, metadata, etc.)
            
        Returns:
            Response from the memory system
            
        Example:
            memory_agent.add("John likes pizza")
            memory_agent.add("Meeting at 3pm", metadata={"type": "appointment"})
        """
        response = self._agent.send_message(
            message=content,
            memorizing=True,
            force_absorb_content=True,
            **kwargs
        )
        return response
    
    def list_users(self) -> Dict[str, Any]:
        """
        List all users in the system.
        
        Returns:
            Dict containing success status, list of users, and any error messages
            
        Example:
            result = memory_agent.list_users()
            if result['success']:
                for user in result['users']:
                    print(f"User: {user['name']} (ID: {user['id']})")
            else:
                print(f"Failed to list users: {result['error']}")
        """
        users = self._agent.client.server.user_manager.list_users()
        return users

    def get_user_by_name(self, user_name: str):
        """
        Get a user by their name.
        
        Args:
            user_name: The name of the user to search for
            
        Returns:
            User object if found, None if not found
            
        Example:
            user = memory_agent.get_user_by_name("Alice")
            if user:
                print(f"Found user: {user.name} (ID: {user.id})")
            else:
                print("User not found")
        """
        users = self.list_users()
        for user in users:
            if user.name == user_name:
                return user
        return None

    def chat(self, message: str, **kwargs) -> str:
        """
        Chat with the memory agent.
        
        Args:
            message: Your message/question
            **kwargs: Additional options
            
        Returns:
            Agent's response
            
        Example:
            response = memory_agent.chat("What does John like?")
            # Returns: "According to my memory, John likes pizza."
        """
        response = self._agent.send_message(
            message=message,
            memorizing=False,  # Chat mode, not memorizing by default
            **kwargs
        )
        # Extract text response
        if isinstance(response, dict):
            return response.get("response", response.get("message", str(response)))
        return str(response)
    
    def clear(self) -> Dict[str, Any]:
        """
        Clear all memories.
        
        Note: This requires manual database file removal and app restart.
        
        Returns:
            Dict with warning message and instructions
            
        Example:
            result = memory_agent.clear()
            print(result['warning'])
            for step in result['instructions']:
                print(step)
        """
        return {
            'success': False,
            'warning': 'Memory clearing requires manual database reset.',
            'instructions': [
                '1. Stop the Mirix application/process',
                '2. Remove the database file: ~/.mirix/sqlite.db',
                '3. Restart the Mirix application',
                '4. Initialize a new Mirix agent'
            ],
            'manual_command': 'rm ~/.mirix/sqlite.db',
            'note': 'After removing the database file, you must restart your application and create a new agent instance.'
        }
    
    def clear_conversation_history(self, user_id: Optional[str] = None) -> Dict[str, Any]:
        """
        Clear conversation history while preserving memories.
        
        This removes user and assistant messages from the conversation
        history but keeps system messages and all stored memories intact.
        
        Args:
            user_id: User ID to clear messages for. If None, clears all messages
                    except system messages. If provided, only clears messages for that specific user.
        
        Returns:
            Dict containing success status, message, and count of deleted messages
            
        Example:
            # Clear all conversation history
            result = memory_agent.clear_conversation_history()
            
            # Clear history for specific user
            result = memory_agent.clear_conversation_history(user_id="user_123")
            
            if result['success']:
                print(f"Cleared {result['messages_deleted']} messages")
            else:
                print(f"Failed to clear: {result['error']}")
        """
        try:
            if user_id is None:
                # Clear all messages except system messages (original behavior)
                current_messages = self._agent.client.server.agent_manager.get_in_context_messages(
                    agent_id=self._agent.agent_states.agent_state.id,
                    actor=self._agent.client.user
                )
                messages_count = len(current_messages)
                
                # Clear conversation history using the agent manager reset_messages method
                self._agent.client.server.agent_manager.reset_messages(
                    agent_id=self._agent.agent_states.agent_state.id,
                    actor=self._agent.client.user,
                    add_default_initial_messages=True  # Keep system message and initial setup
                )
                
                return {
                    'success': True,
                    'message': f"Successfully cleared conversation history. All user and assistant messages removed (system messages preserved).",
                    'messages_deleted': messages_count
                }
            else:
                # Get the user object by ID
                target_user = self._agent.client.server.user_manager.get_user_by_id(user_id)
                if not target_user:
                    return {
                        'success': False,
                        'error': f"User with ID '{user_id}' not found",
                        'messages_deleted': 0
                    }
                
                # Clear messages for specific user (same as FastAPI server implementation)
                # Get current message count for this specific user for reporting
                current_messages = self._agent.client.server.agent_manager.get_in_context_messages(
                    agent_id=self._agent.agent_states.agent_state.id,
                    actor=target_user
                )
                # Count messages belonging to this user (excluding system messages)
                user_messages_count = len([msg for msg in current_messages if msg.role != 'system' and msg.user_id == target_user.id])
                
                # Clear conversation history using the agent manager reset_messages method
                self._agent.client.server.agent_manager.reset_messages(
                    agent_id=self._agent.agent_states.agent_state.id,
                    actor=target_user,
                    add_default_initial_messages=True  # Keep system message and initial setup
                )
                
                return {
                    'success': True,
                    'message': f"Successfully cleared conversation history for {target_user.name}. Messages from other users and system messages preserved.",
                    'messages_deleted': user_messages_count
                }
            
        except Exception as e:
            return {
                'success': False,
                'error': str(e),
                'messages_deleted': 0
            }
    
    def save(self, path: Optional[str] = None) -> Dict[str, Any]:
        """
        Save the current memory state to disk.
        
        Creates a complete backup including agent configuration and database.
        
        Args:
            path: Save directory path (optional). If not provided, generates
                 timestamp-based directory name.
            
        Returns:
            Dict containing success status and backup path
            
        Example:
            result = memory_agent.save("./my_backup")
            if result['success']:
                print(f"Backup saved to: {result['path']}")
            else:
                print(f"Backup failed: {result['error']}")
        """
        from datetime import datetime
        
        if not path:
            path = f"./mirix_backup_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
        
        try:
            result = self._agent.save_agent(path)
            return {
                'success': True,
                'path': path,
                'message': result.get('message', 'Backup completed successfully')
            }
        except Exception as e:
            return {
                'success': False,
                'path': path,
                'error': str(e)
            }
    
    def load(self, path: str) -> Dict[str, Any]:
        """
        Load memory state from a backup directory.
        
        Restores both agent configuration and database from backup.
        
        Args:
            path: Path to backup directory
            
        Returns:
            Dict containing success status and any error messages
            
        Example:
            result = memory_agent.load("./my_backup")
            if result['success']:
                print("Memory restored successfully")
            else:
                print(f"Restore failed: {result['error']}")
        """
        try:
            # result = self._agent.load_agent(path)
            config_path = Path(path) / "mirix_config.yaml"
            self._agent = AgentWrapper(str(config_path), load_from=path)
            return {
                    'success': True,
                    'message': 'Memory state loaded successfully'
                }
        except Exception as e:
            return {
                'success': False,
                'error': str(e)
            }
    
    def _reload_model_settings(self):
        """
        Force reload of model_settings to pick up new environment variables.
        
        This is necessary because Pydantic BaseSettings loads environment variables
        at class instantiation time, which happens at import. Since the SDK sets
        environment variables after import, we need to manually update the singleton.
        """
        from mirix.settings import ModelSettings
        
        # Create a new instance with current environment variables
        new_settings = ModelSettings()
        
        # Update the global singleton instance with new values
        import mirix.settings
        for field_name in ModelSettings.model_fields:
            setattr(mirix.settings.model_settings, field_name, getattr(new_settings, field_name))
    
    def create_user(self, user_name: str) -> Dict[str, Any]:
        """
        Create a new user in the system.
        
        Args:
            name: The name for the new user
            
        Returns:
            Dict containing success status, message, and user data
            
        Example:
            result = memory_agent.create_user("Alice")
        """
        return self._agent.create_user(name=user_name)['user']
    
    def __call__(self, message: str) -> str:
        """
        Allow using the agent as a callable.
        
        Example:
            memory_agent = Mirix(api_key="...")
            response = memory_agent("What do you remember?")
        """
        return self.chat(message)



import re
from typing import List, Optional

from mirix.agent import Agent, AgentState
from mirix.schemas.knowledge_vault import KnowledgeVaultItemBase
from mirix.schemas.episodic_memory import EpisodicEventForLLM
from mirix.schemas.resource_memory import ResourceMemoryItemBase
from mirix.schemas.procedural_memory import ProceduralMemoryItemBase
from mirix.schemas.semantic_memory import SemanticMemoryItemBase

def core_memory_append(self: "Agent", agent_state: "AgentState", label: str, content: str) -> Optional[str]:  # type: ignore
    """
    Append to the contents of core memory. The content will be appended to the end of the block with the given label. If you hit the limit, you can use `core_memory_rewrite` to rewrite the entire block to shorten the content. Note that "Line n:" is only for your visualization of the memory, and you should not include it in the content.

    Args:
        label (str): Section of the memory to be edited (persona or human).
        content (str): Content to write to the memory. All unicode (including emojis) are supported.

    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    # check if the content starts with something like "Line n:" (here n is a number) using regex
    if re.match(r"^Line \d+:", content):
        raise ValueError("You should not include 'Line n:' (here n is a number) in the content.")

    current_value = str(agent_state.memory.get_block(label).value)
    new_value = (current_value + "\n" + str(content)).strip()
    agent_state.memory.update_block_value(label=label, value=new_value)
    return None

def core_memory_rewrite(self: "Agent", agent_state: "AgentState", label: str, content: str) -> Optional[str]:  # type: ignore
    """
    Rewrite the entire content of block <label> in core memory. The entire content in that block will be replaced with the new content. If the old content is full, and you have to rewrite the entire content, make sure to be extremely concise and make it shorter than 20% of the limit.

    Args:
        label (str): Section of the memory to be edited (persona or human).
        content (str): Content to write to the memory. All unicode (including emojis) are supported.
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    current_value = str(agent_state.memory.get_block(label).value)
    new_value = content.strip()
    if current_value != new_value:
        agent_state.memory.update_block_value(label=label, value=new_value)
    return None

def core_memory_replace(self: "Agent", agent_state: "AgentState", label: str, old_line_numbers: str, new_content: str) -> Optional[str]:  # type: ignore
    """
    Replace the contents of core memory. `old_line_numbers` can be a single line number or a range. For instance, it should be "Line 5" to update line 5, or "Line 5-8" to update/delete lines 5 through 8 in the memory. To delete memories, use an empty string for `new_content`. Otherwise, the content will be replaced with the new content. When replacing a range of lines, the new_content will replace all the specified lines.
     
    Args:
        label (str): Section of the memory to be edited (persona or human).
        old_line_numbers (str): The line number(s) of the content to be updated. Can be "Line 5" for single line or "Line 5-8" for a range.
        new_content (str): Content to write to the memory. All unicode (including emojis) are supported. Can contain newlines to create multiple lines.

    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    current_value = str(agent_state.memory.get_block(label).value)
    current_value = current_value.split("\n")
    
    # Parse line numbers
    line_part = old_line_numbers.split(" ")[-1]  # Extract the number part after "Line "
    
    if "-" in line_part:
        # Range of lines
        start_line, end_line = line_part.split("-")
        start_idx = int(start_line) - 1  # Convert to 0-based index
        end_idx = int(end_line) - 1      # Convert to 0-based index
    else:
        # Single line
        start_idx = end_idx = int(line_part) - 1
    
    if new_content != "":
        # Replace the range with new content
        new_lines = new_content.split("\n")
        current_value = current_value[:start_idx] + new_lines + current_value[end_idx + 1:]
    else:
        # Delete the lines
        current_value = current_value[:start_idx] + current_value[end_idx + 1:]
    
    agent_state.memory.update_block_value(label=label, value="\n".join(current_value))
    return None

def episodic_memory_insert(self: "Agent", items: List[EpisodicEventForLLM]):
    """
    The tool to update episodic memory. The item being inserted into the episodic memory is an event either happened on the user or the assistant.

    Args:
        items (array): List of episodic memory items to insert.

    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    for item in items:
        self.episodic_memory_manager.insert_event(
            actor=self.user,
            agent_state=self.agent_state,
            timestamp=item['occurred_at'],
            event_type=item['event_type'],
            event_actor=item['actor'],
            summary=item['summary'],
            details=item['details'],
            organization_id=self.user.organization_id,
            tree_path=item.get('tree_path')
        )
    response = "Events inserted! Now you need to check if there are repeated events shown in the system prompt."
    return response

def episodic_memory_merge(self: "Agent", event_id: str, combined_summary: str = None, combined_details: str = None):
    """
    The tool to merge the new episodic event into the selected episodic event by event_id, should be used when the user is continuing doing the same thing with more details. The combined_summary and combined_details will overwrite the old summary and details of the selected episodic event. Thus DO NOT use "User continues xxx" as the combined_summary because the old one WILL BE OVERWRITTEN and then we can only see "User continus xxx" without the old event.
        
    Args:
        event_id (str): This is the id of which episodic event to append to. 
        combined_summary (str): The updated summary. Note that it will overwrite the old summary so make sure to include the information from the old summary. The new summary needs to be only slightly different from the old summary.
        combined_details (str): The new details to add into the details of the selected episodic event.
    
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """

    episodic_memory = self.episodic_memory_manager.update_event(
        event_id=event_id,
        new_summary=combined_summary,
        new_details=combined_details,
        actor=self.user
    )
    response = "These are the `summary` and the `details` of the updated event:\n", str({'event_id': episodic_memory.id, 'summary': episodic_memory.summary, 'details': episodic_memory.details}) + "\nIf the `details` are too verbose, or the `summary` cannot cover the information in the `details`, call episodic_memory_replace to update this event."
    return response

def episodic_memory_replace(self: "Agent", event_ids: List[str], new_items: List[EpisodicEventForLLM]):
    """
    The tool to replace or delete items in the episodic memory. To replace the memory, set the event_ids to be the ids of the events that needs to be replaced and new_items as the updated events. Note that the number of new items does not need to be the same as the number of event_ids as it is not a one-to-one mapping. To delete the memory, set the event_ids to be the ids of the events that needs to be deleted and new_items as an empty list. To insert new events, use episodic_memory_insert function.

    Args:
        event_ids (str): The ids of the episodic events to be deleted (or replaced).
        new_items (array): List of new episodic memory items to insert. If this is an empty list, then it means that the items are being deleted.
    """

    for event_id in event_ids:
        # It will raise an error if the event_id is not found in the episodic memory.
        self.episodic_memory_manager.get_episodic_memory_by_id(event_id, actor=self.user)

    for event_id in event_ids:
        self.episodic_memory_manager.delete_event_by_id(event_id, actor=self.user)

    for new_item in new_items:
        self.episodic_memory_manager.insert_event(
            actor=self.user,
            agent_state=self.agent_state,
            timestamp=new_item['occurred_at'],
            event_type=new_item['event_type'],
            event_actor=new_item['actor'],
            summary=new_item['summary'],
            details=new_item['details'],
            organization_id=self.user.organization_id,
            tree_path=new_item.get('tree_path')
        )

def check_episodic_memory(self: "Agent", event_ids: List[str], timezone_str: str) -> List[EpisodicEventForLLM]:
    """
    The tool to check the episodic memory. This function will return the episodic events with the given event_ids.

    Args:
        event_ids (str): The ids of the episodic events to be checked.
    
    Returns:
        List[EpisodicEventForLLM]: List of episodic events with the given event_ids.
    """
    episodic_memory = [
        self.episodic_memory_manager.get_episodic_memory_by_id(event_id, timezone_str=timezone_str, actor=self.user) for event_id in event_ids
    ]

    formatted_results = [{'event_id': x.id, 'timestamp': x.occurred_at, 'event_type': x.event_type, 'actor': x.actor, 'summary': x.summary, 'details': x.details, 'tree_path': x.tree_path} for x in episodic_memory]

    return formatted_results

def resource_memory_insert(self: "Agent", items: List[ResourceMemoryItemBase]):
    """
    The tool to insert new items into resource memory.

    Args:
        items (array): List of resource memory items to insert.
    
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """

    for item in items:
        self.resource_memory_manager.insert_resource(
            agent_state=self.agent_state,
            title=item['title'],
            summary=item['summary'],
            resource_type=item['resource_type'],
            content=item['content'],
            actor=self.user,
            organization_id=self.user.organization_id,
            tree_path=item.get('tree_path')
        )

def resource_memory_update(self: "Agent", old_ids: List[str], new_items: List[ResourceMemoryItemBase]):
    """
    The tool to update and delete items in the resource memory. To update the memory, set the old_ids to be the ids of the items that needs to be updated and new_items as the updated items. Note that the number of new items does not need to be the same as the number of old ids as it is not a one-to-one mapping. To delete the memory, set the old_ids to be the ids of the items that needs to be deleted and new_items as an empty list.

    Args:
        old_ids (array): List of ids of the items to be deleted (or updated).
        new_items (array): List of new resource memory items to insert. If this is an empty list, then it means that the items are being deleted.
    """
    
    for old_id in old_ids:
        self.resource_memory_manager.delete_resource_by_id(
            resource_id=old_id,
            actor=self.user
        )
    
    for item in new_items:
        self.resource_memory_manager.insert_resource(
            agent_state=self.agent_state,
            title=item['title'],
            summary=item['summary'],
            resource_type=item['resource_type'],
            content=item['content'],
            actor=self.user,
            organization_id=self.user.organization_id,
            tree_path=item.get('tree_path')
        )

def procedural_memory_insert(self: "Agent", items: List[ProceduralMemoryItemBase]):
    """
    The tool to insert new procedures into procedural memory. Note that the `summary` should not be a general term such as "guide" or "workflow" but rather a more informative description of the procedure.

    Args:
        items (array): List of procedural memory items to insert.
        
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    for item in items:
        self.procedural_memory_manager.insert_procedure(
            agent_state=self.agent_state,
            entry_type=item['entry_type'],
            summary=item['summary'],
            steps=item['steps'],
            actor=self.user,
            organization_id=self.user.organization_id,
            tree_path=item.get('tree_path')
        )

def procedural_memory_update(self: "Agent", old_ids: List[str], new_items: List[ProceduralMemoryItemBase]):
    """
    The tool to update/delete items in the procedural memory. To update the memory, set the old_ids to be the ids of the items that needs to be updated and new_items as the updated items. Note that the number of new items does not need to be the same as the number of old ids as it is not a one-to-one mapping. To delete the memory, set the old_ids to be the ids of the items that needs to be deleted and new_items as an empty list.
    
    Args:
        old_ids (array): List of ids of the items to be deleted (or updated).
        new_items (array): List of new procedural memory items to insert. If this is an empty list, then it means that the items are being deleted.
    
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    for old_id in old_ids:
        self.procedural_memory_manager.delete_procedure_by_id(
            procedure_id=old_id,
            actor=self.user
        )
    
    for item in new_items:
        self.procedural_memory_manager.insert_procedure(
            agent_state=self.agent_state,
            entry_type=item['entry_type'],
            summary=item['summary'],
            steps=item['steps'],
            actor=self.user,
            organization_id=self.user.organization_id,
            tree_path=item.get('tree_path')
        )

def check_semantic_memory(self: "Agent", semantic_item_ids: List[str], timezone_str: str) -> List[SemanticMemoryItemBase]:
    """
    The tool to check the semantic memory. This function will return the semantic memory items with the given ids.

    Args:
        semantic_item_ids (str): The ids of the semantic memory items to be checked.
    
    Returns:
        List[SemanticMemoryItemBase]: List of semantic memory items with the given ids.
    """
    semantic_memory = [
        self.semantic_memory_manager.get_semantic_item_by_id(semantic_memory_id=id, timezone_str=timezone_str, actor=self.user) for id in semantic_item_ids
    ]

    formatted_results = [{'semantic_item_id': x.id, 'name': x.name, 'summary': x.summary, 'details': x.details, 'source': x.source, 'tree_path': x.tree_path} for x in semantic_memory]

    return formatted_results

def semantic_memory_insert(self: "Agent", items: List[SemanticMemoryItemBase]):
    """
    The tool to insert items into semantic memory. 

    Args:
        items (array): List of semantic memory items to insert.

    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    for item in items:
        self.semantic_memory_manager.insert_semantic_item(
            agent_state=self.agent_state,
            name=item['name'],
            summary=item['summary'],
            details=item['details'],
            source=item['source'],
            tree_path=item['tree_path'],
            organization_id=self.user.organization_id,
            actor=self.user
        )

def semantic_memory_update(self: "Agent", old_semantic_item_ids: List[str], new_items: List[SemanticMemoryItemBase]):
    """
    The tool to update/delete items in the semantic memory. To update the memory, set the old_ids to be the ids of the items that needs to be updated and new_items as the updated items. Note that the number of new items does not need to be the same as the number of old ids as it is not a one-to-one mapping. To delete the memory, set the old_ids to be the ids of the items that needs to be deleted and new_items as an empty list.

    Args:
        old_semantic_item_ids (array): List of ids of the items to be deleted (or updated).
        new_items (array): List of new semantic memory items to insert. If this is an empty list, then it means that the items are being deleted.
    
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """

    for old_id in old_semantic_item_ids:
        self.semantic_memory_manager.delete_semantic_item_by_id(
            semantic_memory_id=old_id,
            actor=self.user
        )
    
    new_ids = []
    for item in new_items:
        inserted_item = self.semantic_memory_manager.insert_semantic_item(
            agent_state=self.agent_state,
            name=item['name'],
            summary=item['summary'],
            details=item['details'],
            source=item['source'],
            actor=self.user,
            tree_path=item['tree_path'],
            organization_id=self.user.organization_id
        )
        new_ids.append(inserted_item.id)
    
    message_to_return = "Semantic memory with the following ids have been deleted: " + str(old_semantic_item_ids) + f". New semantic memory items are created: {str(new_ids)}"
    return message_to_return

def knowledge_vault_insert(self: "Agent", items: List[KnowledgeVaultItemBase]):
    """
    The tool to update knowledge vault.

    Args:
        items (array): List of knowledge vault items to insert.
        
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    for item in items:
        self.knowledge_vault_manager.insert_knowledge(
            agent_state=self.agent_state,
            entry_type=item['entry_type'],
            source=item['source'],
            sensitivity=item['sensitivity'],
            secret_value=item['secret_value'],
            caption=item['caption'],
            actor=self.user,
            organization_id=self.user.organization_id
        )

def knowledge_vault_update(self: "Agent", old_ids: List[str], new_items: List[KnowledgeVaultItemBase]):
    """
    The tool to update/delete items in the knowledge vault. To update the knowledge_vault, set the old_ids to be the ids of the items that needs to be updated and new_items as the updated items. Note that the number of new items does not need to be the same as the number of old ids as it is not a one-to-one mapping. To delete the memory, set the old_ids to be the ids of the items that needs to be deleted and new_items as an empty list.
    
    Args:
        old_ids (array): List of ids of the items to be deleted (or updated).
        new_items (array): List of new knowledge vault items to insert. If this is an empty list, then it means that the items are being deleted.
    
    Returns:
        Optional[str]: None is always returned as this function does not produce a response
    """
    for old_id in old_ids:
        self.knowledge_vault_manager.delete_knowledge_by_id(
            knowledge_vault_item_id=old_id,
            actor=self.user
        )
    
    for item in new_items:
        self.knowledge_vault_manager.insert_knowledge(
            agent_state=self.agent_state,
            entry_type=item['entry_type'],
            source=item['source'],
            sensitivity=item['sensitivity'],
            secret_value=item['secret_value'],
            caption=item['caption'],
            actor=self.user,
            organization_id=self.user.organization_id
        )

def trigger_memory_update_with_instruction(self: "Agent", user_message: object, instruction: str, memory_type: str) -> Optional[str]:
    """
    Choose which memory to update. The function will trigger one specific memory agent with the instruction telling the agent what to do.

    Args:
        instruction (str): The instruction to the memory agent.
        memory_type (str): The type of memory to update. It should be chosen from the following: "core", "episodic", "resource", "procedural", "knowledge_vault", "semantic". For instance, ['episodic', 'resource'].
    
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """

    from mirix import create_client

    client = create_client()
    agents = client.list_agents()

    # Validate that user_message is a dictionary
    if not isinstance(user_message, dict):
        raise TypeError(f"user_message must be a dictionary, got {type(user_message).__name__}: {user_message}")

    # Fallback to sequential processing for backward compatibility
    response = ''

    if memory_type == "core":
        agent_type = "core_memory_agent"
    elif memory_type == "episodic":
        agent_type = "episodic_memory_agent"
    elif memory_type == "resource":
        agent_type = "resource_memory_agent"
    elif memory_type == "procedural":
        agent_type = "procedural_memory_agent"
    elif memory_type == "knowledge_vault":
        agent_type = "knowledge_vault_agent"
    elif memory_type == 'semantic':
        agent_type = "semantic_memory_agent"
    else:
        raise ValueError(f"Memory type '{memory_type}' is not supported. Please choose from 'core', 'episodic', 'resource', 'procedural', 'knowledge_vault', 'semantic'.")

    matching_agent = None
    for agent in agents:
        if agent.agent_type == agent_type:
            matching_agent = agent
            break
    
    if matching_agent is None:
        raise ValueError(f"No agent found with type '{agent_type}'")
    
    client.send_message(
        role='user', 
        user_id=self.user.id,
        agent_id=matching_agent.id, 
        message="[Message from Chat Agent (Now you are allowed to make multiple function calls sequentially)] " + instruction, 
        existing_file_uris=user_message['existing_file_uris'],
        retrieved_memories=user_message.get('retrieved_memories', None)
    )
    response += '[System Message] Agent ' + matching_agent.name + ' has been triggered to update the memory.\n'

    return response.strip()

def trigger_memory_update(self: "Agent", user_message: object, memory_types: List[str]) -> Optional[str]:
    """
    Choose which memory to update. This function will trigger another memory agent which is specifically in charge of handling the corresponding memory to update its memory. Trigger all necessary memory updates at once. Put the explanations in the `internal_monologue` field.

    Args:
        memory_types (List[str]): The types of memory to update. It should be chosen from the following: "core", "episodic", "resource", "procedural", "knowledge_vault", "semantic". For instance, ['episodic', 'resource'].
        
    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """

    from mirix import create_client

    client = create_client()
    agents = client.list_agents()

    # Validate that user_message is a dictionary
    if not isinstance(user_message, dict):
        raise TypeError(f"user_message must be a dictionary, got {type(user_message).__name__}: {user_message}")

    if 'message_queue' in user_message:
        
        from concurrent.futures import ThreadPoolExecutor, as_completed
        from tqdm import tqdm
        import time

        # Use multi-processing approach similar to _send_to_memory_agents_separately
        message_queue = user_message['message_queue']
        
        # Map memory types to agent types
        memory_type_to_agent_type = {
            "core": "core_memory_agent",
            "episodic": "episodic_memory_agent", 
            "resource": "resource_memory_agent",
            "procedural": "procedural_memory_agent",
            "knowledge_vault": "knowledge_vault_agent",
            "semantic": "semantic_memory_agent"
        }
        
        # Filter to only supported memory types
        valid_agent_types = []
        for memory_type in memory_types:
            if memory_type in memory_type_to_agent_type:
                valid_agent_types.append(memory_type_to_agent_type[memory_type])
            else:
                raise ValueError(f"Memory type '{memory_type}' is not supported. Please choose from 'core', 'episodic', 'resource', 'procedural', 'knowledge_vault', 'semantic'.")
        
        if user_message['message'][-1]['type'] == 'text' and user_message['message'][-1]['text'].startswith('[System Message]'):
            user_message['message'][-1]['text'] = "[System Message] Interpret the provided content, extract the important information matching your memory type and save it into the memory."

        # Prepare payloads for message queue
        payloads = {
            'user_id': self.user.id,
            'message': user_message['message'],
            'existing_file_uris': user_message.get('existing_file_uris', set()),
            'chaining': user_message.get('chaining', False),
            'message_queue': message_queue,
            'retrieved_memories': user_message.get('retrieved_memories', None)
        }

        responses = []
        overall_start = time.time()

        # Use ThreadPoolExecutor for parallel processing
        with ThreadPoolExecutor(max_workers=len(valid_agent_types)) as pool:
            futures = []
            for agent_type in valid_agent_types:
                matching_agents = [agent for agent in agents if agent.agent_type == agent_type]
                if not matching_agents:
                    raise ValueError(f"No agent found with type '{agent_type}'")
                futures.append(
                    pool.submit(message_queue.send_message_in_queue, 
                               client, matching_agents[0].id, payloads, agent_type)
                )
            
            for future in tqdm(as_completed(futures), total=len(futures)):
                response, agent_type = future.result()
                responses.append(response)

        overall_end = time.time()
        response_message = f'[System Message] {len(valid_agent_types)} memory agents have been triggered in parallel to update the memory. Total time: {overall_end - overall_start:.2f} seconds.'
        return response_message

    else:
        
        # Fallback to sequential processing for backward compatibility
        response = ''

        for memory_type in memory_types:

            if memory_type == "core":
                agent_type = "core_memory_agent"
            elif memory_type == "episodic":
                agent_type = "episodic_memory_agent"
            elif memory_type == "resource":
                agent_type = "resource_memory_agent"
            elif memory_type == "procedural":
                agent_type = "procedural_memory_agent"
            elif memory_type == "knowledge_vault":
                agent_type = "knowledge_vault_agent"
            elif memory_type == 'semantic':
                agent_type = "semantic_memory_agent"
            else:
                raise ValueError(f"Memory type '{memory_type}' is not supported. Please choose from 'core', 'episodic', 'resource', 'procedural', 'knowledge_vault', 'semantic'.")

            matching_agent = None
            for agent in agents:
                if agent.agent_type == agent_type:
                    matching_agent = agent
                    break
            
            if matching_agent is None:
                raise ValueError(f"No agent found with type '{agent_type}'")
            
            client.send_message(agent_id=matching_agent.id, 
                role='user', 
                message=user_message['message'], 
                existing_file_uris=user_message['existing_file_uris'],
                retrieved_memories=user_message.get('retrieved_memories', None)
            )
            response += '[System Message] Agent ' + matching_agent.name + ' has been triggered to update the memory.\n'
        
        return response.strip()

def finish_memory_update(self: "Agent"):
    """
    Finish the memory update process. This function should be called after the Memory is updated.

    Returns:
        Optional[str]: None is always returned as this function does not produce a response.
    """
    return None


import base64
import io

# Convert images to base64
def encode_image(image_path):
    with open(image_path, "rb") as img_file:
        return base64.b64encode(img_file.read()).decode("utf-8")

def encode_image_from_pil(image):
    buffer = io.BytesIO()
    image.save(buffer, format="PNG")  # Change format if needed (JPEG, etc.)
    return base64.b64encode(buffer.getvalue()).decode("utf-8")


from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field


class SystemMessage(BaseModel):
    content: str
    role: str = "system"
    name: Optional[str] = None


class UserMessage(BaseModel):
    content: Union[str, List[dict]]
    role: str = "user"
    name: Optional[str] = None


class ToolCallFunction(BaseModel):
    name: str
    arguments: str


class ToolCall(BaseModel):
    id: str
    type: Literal["function"] = "function"
    function: ToolCallFunction


class AssistantMessage(BaseModel):
    content: Optional[str] = None
    role: str = "assistant"
    name: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None


class ToolMessage(BaseModel):
    content: str
    role: str = "tool"
    tool_call_id: str


ChatMessage = Union[SystemMessage, UserMessage, AssistantMessage, ToolMessage]


# TODO: this might not be necessary with the validator
def cast_message_to_subtype(m_dict: dict) -> ChatMessage:
    """Cast a dictionary to one of the individual message types"""
    role = m_dict.get("role")
    if role == "system" or role == "developer":
        return SystemMessage(**m_dict)
    elif role == "user":
        return UserMessage(**m_dict)
    elif role == "assistant":
        return AssistantMessage(**m_dict)
    elif role == "tool":
        return ToolMessage(**m_dict)
    else:
        raise ValueError(f"Unknown message role: {role}")


class ResponseFormat(BaseModel):
    type: str = Field(default="text", pattern="^(text|json_object)$")


## tool_choice ##
class FunctionCall(BaseModel):
    name: str


class ToolFunctionChoice(BaseModel):
    # The type of the tool. Currently, only function is supported
    type: Literal["function"] = "function"
    # type: str = Field(default="function", const=True)
    function: FunctionCall


ToolChoice = Union[Literal["none", "auto", "required"], ToolFunctionChoice]


## tools ##
class FunctionSchema(BaseModel):
    name: str
    description: Optional[str] = None
    parameters: Optional[Dict[str, Any]] = None  # JSON Schema for the parameters


class Tool(BaseModel):
    # The type of the tool. Currently, only function is supported
    type: Literal["function"] = "function"
    # type: str = Field(default="function", const=True)
    function: FunctionSchema


## function_call ##
FunctionCallChoice = Union[Literal["none", "auto"], FunctionCall]


class ChatCompletionRequest(BaseModel):
    """https://platform.openai.com/docs/api-reference/chat/create"""

    model: str
    messages: List[ChatMessage]
    frequency_penalty: Optional[float] = 0
    logit_bias: Optional[Dict[str, int]] = None
    logprobs: Optional[bool] = False
    top_logprobs: Optional[int] = None
    max_tokens: Optional[int] = None
    n: Optional[int] = 1
    presence_penalty: Optional[float] = 0
    response_format: Optional[ResponseFormat] = None
    seed: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    stream: Optional[bool] = False
    temperature: Optional[float] = 1 # TODO: might need to add logics to control this
    top_p: Optional[float] = 1

    # function-calling related
    tools: Optional[List[Tool]] = None
    tool_choice: Optional[ToolChoice] = None  # "none" means don't call a tool
    # deprecated scheme
    functions: Optional[List[FunctionSchema]] = None
    function_call: Optional[FunctionCallChoice] = None



from typing import TYPE_CHECKING, List

from sqlalchemy.orm import Mapped, mapped_column, relationship

from mirix.orm.mixins import OrganizationMixin
from mirix.orm.sqlalchemy_base import SqlalchemyBase
from mirix.schemas.user import User as PydanticUser

if TYPE_CHECKING:
    from mirix.orm import Job, Organization


class User(SqlalchemyBase, OrganizationMixin):
    """User ORM class"""

    __tablename__ = "users"
    __pydantic_model__ = PydanticUser

    name: Mapped[str] = mapped_column(nullable=False, doc="The display name of the user.")
    status: Mapped[str] = mapped_column(nullable=False, doc="Whether the user is active or not.")
    timezone: Mapped[str] = mapped_column(nullable=False, doc="The timezone of the user.")

    # relationships
    organization: Mapped["Organization"] = relationship("Organization", back_populates="users")

    # TODO: Add this back later potentially
    # tokens: Mapped[List["Token"]] = relationship("Token", back_populates="user", doc="the tokens associated with this user.")



import uuid
import time
import threading
import traceback


class MessageQueue:
    """
    Handles queueing and ordering of messages to different agent types.
    Ensures that messages of the same type are processed in order.
    """
    
    def __init__(self):
        self.message_queue = {}
        self._message_queue_lock = threading.Lock()
    
    def send_message_in_queue(self, client, agent_id, kwargs, agent_type='chat'):
        """
        Queue a message to be sent to a specific agent type.
        
        Args:
            client: The mirix client instance
            agent_id: The ID of the agent to send the message to
            kwargs: Arguments to pass to client.send_message
            agent_type: Type of agent to send message to
            
        Returns:
            Tuple of (response, agent_type)
        """
        message_uuid = uuid.uuid4()

        with self._message_queue_lock:
            self.message_queue[message_uuid] = {
                'kwargs': kwargs,
                'started': False,
                'finished': False,
                'type': agent_type,
            }

        # Wait for earlier requests of the same type to finish
        while not self._check_if_earlier_requests_are_finished(message_uuid):
            time.sleep(0.1)

        with self._message_queue_lock:
            self.message_queue[message_uuid]['started'] = True

        try:
            response = client.send_message(
                agent_id=agent_id,
                role='user',
                **self.message_queue[message_uuid]['kwargs']
            )
        except Exception as e:
            print(f"Error sending message: {e}")
            print(traceback.format_exc())
            print("agent_type: ", agent_type, "gets error. agent_id: ", agent_id, "ERROR")
            response = "ERROR"

        with self._message_queue_lock:
            self.message_queue[message_uuid]['finished'] = True
            del self.message_queue[message_uuid]
        
        return response, agent_type
    
    def _check_if_earlier_requests_are_finished(self, message_uuid):
        """Check if all earlier requests of the same type have finished."""
        with self._message_queue_lock:
            if message_uuid not in self.message_queue:
                raise ValueError("Message not found in the queue.")
            
            # Get current message type
            current_message_type = self.message_queue[message_uuid]['type']
            
            # Find index of current message
            message_keys = list(self.message_queue.keys())
            idx = message_keys.index(message_uuid)
            
            # Check earlier messages of the same type
            for i in range(idx):
                earlier_message = self.message_queue[message_keys[i]]
                if earlier_message['type'] == current_message_type:
                    if not earlier_message['finished']:
                        return False
            
            return True
    
    def _get_agent_id_for_type(self, agent_states, agent_type):
        """Get the agent ID for the specified agent type."""
        agent_type_to_state_mapping = {
            'chat': 'agent_state',
            'episodic_memory': 'episodic_memory_agent_state',
            'procedural_memory': 'procedural_memory_agent_state',
            'knowledge_vault': 'knowledge_vault_agent_state',
            'meta_memory': 'meta_memory_agent_state',
            'semantic_memory': 'semantic_memory_agent_state',
            'core_memory': 'core_memory_agent_state',
            'resource_memory': 'resource_memory_agent_state',
            'meta_memory_agent': 'meta_memory_agent_state',  # Alias
        }
        
        state_name = agent_type_to_state_mapping.get(agent_type)
        if not state_name:
            raise ValueError(f"Unknown agent type: {agent_type}")
        
        if not hasattr(agent_states, state_name):
            raise ValueError(f"Agent state {state_name} not found")

        return getattr(agent_states, state_name).id
    
    def get_queue_length(self):
        """Get the current length of the message queue."""
        with self._message_queue_lock:
            return len(self.message_queue)



from typing import Any, Dict, List, Optional

from pydantic import Field, model_validator

from mirix.constants import (
    COMPOSIO_TOOL_TAG_NAME,
    FUNCTION_RETURN_CHAR_LIMIT,
    MIRIX_CORE_TOOL_MODULE_NAME,
    MIRIX_MEMORY_TOOL_MODULE_NAME,
    MIRIX_EXTRA_TOOL_MODULE_NAME,
)
from mirix.functions.functions import derive_openai_json_schema, get_json_schema_from_module
from mirix.functions.helpers import generate_langchain_tool_wrapper
from mirix.functions.schema_generator import generate_schema_from_args_schema_v2
from mirix.orm.enums import ToolType
from mirix.schemas.mirix_base import MirixBase


class BaseTool(MirixBase):
    __id_prefix__ = "tool"


class Tool(BaseTool):
    """
    Representation of a tool, which is a function that can be called by the agent.

    Parameters:
        id (str): The unique identifier of the tool.
        name (str): The name of the function.
        tags (List[str]): Metadata tags.
        source_code (str): The source code of the function.
        json_schema (Dict): The JSON schema of the function.

    """

    id: str = BaseTool.generate_id_field()
    tool_type: ToolType = Field(ToolType.CUSTOM, description="The type of the tool.")
    description: Optional[str] = Field(None, description="The description of the tool.")
    source_type: Optional[str] = Field(None, description="The type of the source code.")
    organization_id: Optional[str] = Field(None, description="The unique identifier of the organization associated with the tool.")
    name: Optional[str] = Field(None, description="The name of the function.")
    tags: List[str] = Field([], description="Metadata tags.")

    # code
    source_code: Optional[str] = Field(None, description="The source code of the function.")
    json_schema: Optional[Dict] = Field(None, description="The JSON schema of the function.")

    # tool configuration
    return_char_limit: int = Field(FUNCTION_RETURN_CHAR_LIMIT, description="The maximum number of characters in the response.")

    # metadata fields
    created_by_id: Optional[str] = Field(None, description="The id of the user that made this Tool.")
    last_updated_by_id: Optional[str] = Field(None, description="The id of the user that made this Tool.")

    @model_validator(mode="after")
    def populate_missing_fields(self):
        """
        Populate missing fields: name, description, and json_schema.
        """

        if self.tool_type == ToolType.CUSTOM:
            # If it's a custom tool, we need to ensure source_code is present
            if not self.source_code:
                raise ValueError(f"Custom tool with id={self.id} is missing source_code field.")

            # Always derive json_schema for freshest possible json_schema
            # TODO: Instead of checking the tag, we should having `COMPOSIO` as a specific ToolType
            # TODO: We skip this for Composio bc composio json schemas are derived differently
            if not (COMPOSIO_TOOL_TAG_NAME in self.tags):
                self.json_schema = derive_openai_json_schema(source_code=self.source_code)
        elif self.tool_type in {ToolType.MIRIX_CORE}:
            # If it's mirix core tool, we generate the json_schema on the fly here
            self.json_schema = get_json_schema_from_module(module_name=MIRIX_CORE_TOOL_MODULE_NAME, function_name=self.name)
        elif self.tool_type in {ToolType.MIRIX_MEMORY_CORE}:
            self.json_schema = get_json_schema_from_module(module_name=MIRIX_MEMORY_TOOL_MODULE_NAME, function_name=self.name)
        elif self.tool_type in {ToolType.MIRIX_EXTRA}:
            self.json_schema = get_json_schema_from_module(module_name=MIRIX_EXTRA_TOOL_MODULE_NAME, function_name=self.name)
        elif self.tool_type in {ToolType.MIRIX_MCP}:
            # MCP tools have their json_schema already provided by MCP tool registry
            # Skip validation since these are auto-generated tools
            if not self.json_schema:
                raise ValueError(f"MCP tool {self.name} is missing json_schema field")

        # Derive name from the JSON schema if not provided
        if not self.name:
            # TODO: This in theory could error, but name should always be on json_schema
            # TODO: Make JSON schema a typed pydantic object
            self.name = self.json_schema.get("name")

        # Derive description from the JSON schema if not provided
        if not self.description:
            # TODO: This in theory could error, but description should always be on json_schema
            # TODO: Make JSON schema a typed pydantic object
            self.description = self.json_schema.get("description")

        return self


class ToolCreate(MirixBase):
    name: Optional[str] = Field(None, description="The name of the function (auto-generated from source_code if not provided).")
    description: Optional[str] = Field(None, description="The description of the tool.")
    tags: List[str] = Field([], description="Metadata tags.")
    source_code: str = Field(..., description="The source code of the function.")
    source_type: str = Field("python", description="The source type of the function.")
    json_schema: Optional[Dict] = Field(
        None, description="The JSON schema of the function (auto-generated from source_code if not provided)"
    )
    return_char_limit: int = Field(FUNCTION_RETURN_CHAR_LIMIT, description="The maximum number of characters in the response.")

    @classmethod
    def from_langchain(
        cls,
        langchain_tool: "LangChainBaseTool",
        additional_imports_module_attr_map: dict[str, str] = None,
    ) -> "ToolCreate":
        """
        Class method to create an instance of Tool from a Langchain tool (must be from langchain_community.tools).

        Args:
            langchain_tool (LangChainBaseTool): An instance of a LangChain BaseTool (BaseTool from LangChain)
            additional_imports_module_attr_map (dict[str, str]): A mapping of module names to attribute name. This is used internally to import all the required classes for the langchain tool. For example, you would pass in `{"langchain_community.utilities": "WikipediaAPIWrapper"}` for `from langchain_community.tools import WikipediaQueryRun`. NOTE: You do NOT need to specify the tool import here, that is done automatically for you.

        Returns:
            Tool: A Mirix Tool initialized with attributes derived from the provided LangChain BaseTool object.
        """
        description = langchain_tool.description
        source_type = "python"
        tags = ["langchain"]
        # NOTE: langchain tools may come from different packages
        wrapper_func_name, wrapper_function_str = generate_langchain_tool_wrapper(langchain_tool, additional_imports_module_attr_map)
        json_schema = generate_schema_from_args_schema_v2(langchain_tool.args_schema, name=wrapper_func_name, description=description)

        return cls(
            name=wrapper_func_name,
            description=description,
            source_type=source_type,
            tags=tags,
            source_code=wrapper_function_str,
            json_schema=json_schema,
        )

    @classmethod
    def load_default_langchain_tools(cls) -> List["ToolCreate"]:
        # For now, we only support wikipedia tool
        from langchain_community.tools import WikipediaQueryRun
        from langchain_community.utilities import WikipediaAPIWrapper

        wikipedia_tool = ToolCreate.from_langchain(
            WikipediaQueryRun(api_wrapper=WikipediaAPIWrapper()), {"langchain_community.utilities": "WikipediaAPIWrapper"}
        )

        return [wikipedia_tool]

    @classmethod
    def load_default_composio_tools(cls) -> List["ToolCreate"]:
        pass

        # TODO: Disable composio tools for now
        # TODO: Naming is causing issues
        # calculator = ToolCreate.from_composio(action_name=Action.MATHEMATICAL_CALCULATOR.name)
        # serp_news = ToolCreate.from_composio(action_name=Action.SERPAPI_NEWS_SEARCH.name)
        # serp_google_search = ToolCreate.from_composio(action_name=Action.SERPAPI_SEARCH.name)
        # serp_google_maps = ToolCreate.from_composio(action_name=Action.SERPAPI_GOOGLE_MAPS_SEARCH.name)

        return []


class ToolUpdate(MirixBase):
    description: Optional[str] = Field(None, description="The description of the tool.")
    name: Optional[str] = Field(None, description="The name of the function.")
    tags: Optional[List[str]] = Field(None, description="Metadata tags.")
    source_code: Optional[str] = Field(None, description="The source code of the function.")
    source_type: Optional[str] = Field(None, description="The type of the source code.")
    json_schema: Optional[Dict] = Field(
        None, description="The JSON schema of the function (auto-generated from source_code if not provided)"
    )
    return_char_limit: Optional[int] = Field(None, description="The maximum number of characters in the response.")

    class Config:
        extra = "ignore"  # Allows extra fields without validation errors
        # TODO: Remove this, and clean usage of ToolUpdate everywhere else


class ToolRunFromSource(MirixBase):
    source_code: str = Field(..., description="The source code of the function.")
    args: Dict[str, Any] = Field(..., description="The arguments to pass to the tool.")
    env_vars: Dict[str, str] = Field(None, description="The environment variables to pass to the tool.")
    name: Optional[str] = Field(None, description="The name of the tool to run.")
    source_type: Optional[str] = Field(None, description="The type of the source code.")
